// Portions of this file taken from
// Petko Manolov - Petkan (petkan@dce.bg)
// from his driver pegasus.h

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifdef	DEBUG
#ifdef	DEVELOPER_DEBUG
#define vkg_printk                         	printk
#else
#define	vkg_printk							dbg
#endif
#else
#define	vkg_printk(...)
#endif

/* From CDC Spec Table 24 */
#define CS_INTERFACE			0x24

#define GS_VENDID_MODEM 0x1110
#define GS_PRODID_MODEM 0x5C01

#define	CDC_ETHER_MAX_MTU		2048

#define	CDC_ETHER_MAX_INTR_PKT	32

//
// This is the MAX sequence count for the Tx payload packets.
#define  MAX_SEQ_COUNT    	16
#define  MAX_PAYLOAD_LEN  	62
#define  VKG_HDR_LEN      	2
#define  MAX_PKT_LEN      	(MAX_PAYLOAD_LEN + VKG_HDR_LEN)

/* These definitions are used with the ether_dev_t flags element */
#define	CDC_ETHER_PRESENT		0x00000001
#define	CDC_ETHER_RUNNING		0x00000002
#define	CDC_ETHER_RX_BUSY		0x00000004
#define	CDC_ETHER_UNPLUG		0x00000008
#define	CDC_ETHER_RX_ISSUED		0x00000010
#define	CDC_ETHER_TX_ISSUED		0x00000020

#define	CDC_ETHER_TX_TIMEOUT	(HZ*10)

#define	TX_UNDERRUN			0x80
#define	EXCESSIVE_COL			0x40
#define	LATE_COL			0x20
#define	NO_CARRIER			0x10
#define	LOSS_CARRIER			0x08
#define	JABBER_TIMEOUT			0x04

#define	CDC_ETHER_REQT_READ		0xc0
#define	CDC_ETHER_REQT_WRITE	0x40
#define	CDC_ETHER_REQ_GET_REGS	0xf0
#define	CDC_ETHER_REQ_SET_REGS	0xf1
#define	CDC_ETHER_REQ_SET_REG	PIPERIDER_REQ_SET_REGS


#define CDC_NOTIFICATION_NETWORK_CONNECTION     	0x00
#define CDC_NETWORK_CONNECTION_DISCONNECT    		0x0000
#define CDC_NETWORK_CONNECTION_CONNECT        		0x0001
#define CDC_NOTIFICATION_CONNECTION_SPEED_CHANGE    0x2A
#define CDC_CONNECTION_SPEED_CHANGE_MSG_SIZE    	16
#define CDC_MINIMUM_NOTIFICATION_MSG_SIZE         8

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,20)
typedef  struct  usb_ctrlrequest USB_CTRLREQUEST;
#define USB_CTRLREQUEST_REQTTYPE(usb_ctrlreq)    ((usb_ctrlreq)->bRequestType)
#define USB_CTRLREQUEST_REQUEST(usb_ctrlreq)    ((usb_ctrlreq)->bRequest)
#define USB_CTRLREQUEST_VALUE(usb_ctrlreq)    ((usb_ctrlreq)->wValue)
#define USB_CTRLREQUEST_INDEX(usb_ctrlreq)    ((usb_ctrlreq)->wIndex)
#define USB_CTRLREQUEST_LENGTH(usb_ctrlreq)    ((usb_ctrlreq)->wLength)
#else
typedef devrequest  USB_CTRLREQUEST;
#define USB_CTRLREQUEST_REQTTYPE(usb_ctrlreq)    ((usb_ctrlreq)->requesttype)
#define USB_CTRLREQUEST_REQUEST(usb_ctrlreq)    ((usb_ctrlreq)->request)
#define USB_CTRLREQUEST_VALUE(usb_ctrlreq)    ((usb_ctrlreq)->value)
#define USB_CTRLREQUEST_INDEX(usb_ctrlreq)    ((usb_ctrlreq)->index)
#define USB_CTRLREQUEST_LENGTH(usb_ctrlreq)    ((usb_ctrlreq)->length)
#endif

typedef  struct _ether_dev_t {
	struct usb_device	*usb;
	struct net_device	*net;
	struct net_device_stats	stats;
	unsigned	flags;
	int			configuration_num;
	int			bConfigurationValue;
	int			comm_interface;
	int			comm_bInterfaceNumber;
	int			comm_interface_altset_num;
	int			comm_bAlternateSetting;
	int			comm_ep_in;
	int			data_interface;
	int			data_bInterfaceNumber;
	int			data_interface_altset_num_with_traffic;
	int			data_bAlternateSetting_with_traffic;
	int			data_interface_altset_num_without_traffic;
	int			data_bAlternateSetting_without_traffic;
	int			intr_ep;
	int			data_ep_in;
	int			data_ep_out;
	int			data_ep_out_size;
	int			intr_ep_size;
	__u16			bcdCDC;
	__u8			iMACAddress;
	__u32			bmEthernetStatistics;
	__u16			wMaxSegmentSize;
	__u16			wNumberMCFilters;
	__u8			bNumberPowerFilters;
	__u16			mode_flags;
	int				intr_interval;
	USB_CTRLREQUEST ctrl_dr;
    int				ctrl_pending;
	struct urb		intr_urb, rx_urb, tx_urb, ctrl_urb;
    int  			txsequence;
	unsigned char   rxsequence;
	unsigned char   rx_out_of_seq;
	unsigned char	rx_buff[CDC_ETHER_MAX_MTU] __attribute__((aligned(L1_CACHE_BYTES)));
	unsigned char	tx_buff[CDC_ETHER_MAX_MTU] __attribute__((aligned(L1_CACHE_BYTES)));
	unsigned char	intr_buff[CDC_ETHER_MAX_INTR_PKT] __attribute__((aligned(L1_CACHE_BYTES))) ;
} ether_dev_t;


typedef struct cdc_notification
{
	unsigned char	bmRequestType;
	unsigned char	bNotification;
	__u16			wValue;
	__u16			wIndex;
	__u16			wLength;
	unsigned char	Data[1];
}cdc_notification_t;


typedef struct cdc_speedchange
{
	__u32		UpStreamBitRate;
	__u32		DownStreamBitRate;	
}cdc_speedchange_t;


/* These definitions used in the Ethernet Packet Filtering requests */
/* See CDC Spec Table 62 */
#define MODE_FLAG_PROMISCUOUS   (1<<0)
#define MODE_FLAG_ALL_MULTICAST (1<<1)
#define MODE_FLAG_DIRECTED      (1<<2)
#define MODE_FLAG_BROADCAST     (1<<3)
#define MODE_FLAG_MULTICAST     (1<<4)

/* CDC Spec class requests - CDC Spec Table 46 */
#define SET_ETHERNET_MULTICAST_FILTER    0x40
#define SET_ETHERNET_PACKET_FILTER       0x43
#define	GET_ETHERNET_STATISTIC			 0x44


/* These definitions are used with the ether_dev_t properties field */
#define HAVE_NOTIFICATION_ELEMENT	0x0001
#define PERFECT_FILTERING		0x0002
#define NO_SET_MULTICAST		0x0004

/* These definitions are used in the requirements parser */
#define REQ_HDR_FUNC_DESCR	0x0001
#define REQ_UNION_FUNC_DESCR	0x0002
#define REQ_ETH_FUNC_DESCR	0x0004
//#define REQUIREMENTS_TOTAL	REQ_ETH_FUNC_DESCR | REQ_UNION_FUNC_DESCR | REQ_HDR_FUNC_DESCR
#define REQUIREMENTS_TOTAL	REQ_ETH_FUNC_DESCR 

/* Some useful lengths */
#define HEADER_FUNC_DESC_LEN	0x5
#define UNION_FUNC_DESC_LEN	0x5
#define ETHERNET_FUNC_DESC_LEN	0xD /* 13 for all you decimal weenies */

#define	VKG_QUICK_SYNC_COUNT	20
#define	FALSE	0
#define	TRUE	1


